import asyncio
from langchain.tools import tool
from functools import wraps
from typing import Optional


def flotool(
    name: str,
    description: Optional[str] = None,
    argument_contract: Optional[type] = None,
    unsafe: bool = False,
):
    def decorator(func):
        func.__doc__ = func.__doc__ or description

        @tool(name, args_schema=argument_contract)
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if unsafe:
                    raise e
                return f'An error occurred while executing the tool: {str(e)}, please retry with the corresponding fix'

        @tool(name, args_schema=argument_contract)
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                if unsafe:
                    raise e
                return f'An error occurred while executing the tool: {str(e)}, please retry with the corresponding fix'

        return async_wrapper if asyncio.iscoroutinefunction(func) else wrapper

    return decorator



from flo_ai.models.flo_team import FloTeam
from flo_ai.yaml.config import (
    FloRoutedTeamConfig,
    TeamConfig,
    AgentConfig,
    FloAgentConfig,
)
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.state.flo_session import FloSession
from flo_ai.router.flo_router_factory import FloRouterFactory
from flo_ai.factory.agent_factory import AgentFactory
from flo_ai.error.flo_exception import FloException
from flo_ai.yaml.validators import raise_for_name_error
from flo_ai.common.flo_logger import get_logger


def build_supervised_team(session: FloSession) -> ExecutableFlo:
    name_set = set()
    flo_config = session.config
    if isinstance(flo_config, FloRoutedTeamConfig):
        team_config: TeamConfig = flo_config.team
        team = parse_and_build_subteams(session, team_config, name_set)
        return team
    elif isinstance(flo_config, FloAgentConfig):
        agent_config: AgentConfig = flo_config.agent
        validate_names(name_set, agent_config.name, session)
        agent = AgentFactory.create(session, agent_config)
        return agent


def validate_team(name_set: set, team_config: TeamConfig, session: FloSession):
    validate_names(name_set, team_config.name, session)


def parse_and_build_subteams(
    session: FloSession, team_config: TeamConfig, name_set=set()
) -> ExecutableFlo:
    flo_team = None
    validate_team(name_set, team_config, session)
    if team_config.agents:
        members = [AgentFactory.create(session, agent) for agent in team_config.agents]
        flo_team = FloTeam.Builder(team_config, members=members).build()
        router = FloRouterFactory.create(session, team_config, flo_team)
        flo_routed_team = router.build_routed_team()
    else:
        flo_teams = []
        for subteam in team_config.subteams:
            flo_subteam = parse_and_build_subteams(session, subteam, name_set)
            flo_teams.append(flo_subteam)
        flo_team = FloTeam.Builder(team_config, members=flo_teams).build()
        router = FloRouterFactory.create(session, team_config, flo_team)
        flo_routed_team = router.build_routed_team()
    return flo_routed_team


def validate_names(name_set: set, name, session: FloSession):
    raise_for_name_error(name)
    if name in name_set:
        get_logger().error(f"Duplicate name found: '{name}'", session)
        raise FloException(
            f"The name '{name}' is duplicate in the config. Make sure all teams and agents have unique names"
        )
    name_set.add(name)



class FloException(Exception):
    def __init__(self, message: str, error_code: int = -1):
        """
        Initialize the FloException with a message and optional error code.

        :param message: Error message to be displayed.
        :param error_code: Optional error code to be associated with the exception.
        """
        self.message = message
        self.error_code = error_code
        super().__init__(message)

    def __str__(self):
        if self.error_code is not None:
            return f'[Error {self.error_code}] {self.message}'
        return self.message



from typing import Union
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable
from flo_ai.state.flo_session import FloSession
from flo_ai.constants.prompt_constants import FLO_FINISH
from flo_ai.router.flo_router import FloRouter
from flo_ai.models.flo_team import FloTeam
from flo_ai.models.flo_routed_team import FloRoutedTeam
from langgraph.graph import StateGraph
from flo_ai.state.flo_state import TeamFloAgentState
from flo_ai.yaml.config import TeamConfig
from langchain_core.output_parsers import JsonOutputParser
from pydantic import BaseModel, Field


class NextAgent(BaseModel):
    next: str = Field(description='Name of the next member to be called')


class FloLLMRouter(FloRouter):
    def __init__(
        self, session: FloSession, executor: Runnable, flo_team: FloTeam, name: str
    ) -> None:
        super().__init__(
            session=session, name=name, flo_team=flo_team, executor=executor
        )

    def build_graph(self):
        flo_agent_nodes = [self.build_node(flo_agent) for flo_agent in self.members]
        workflow = StateGraph(TeamFloAgentState)
        for flo_agent_node in flo_agent_nodes:
            workflow.add_node(flo_agent_node.name, flo_agent_node.func)
        workflow.add_node(self.router_name, self.build_node(self).func)
        for member in self.member_names:
            workflow.add_edge(member, self.router_name)
        workflow.add_conditional_edges(self.router_name, self.router_fn)
        workflow.set_entry_point(self.router_name)
        workflow_graph = workflow.compile()
        return FloRoutedTeam(self.flo_team.name, workflow_graph, self.flo_team.config)

    class Builder:
        def __init__(
            self,
            session: FloSession,
            team_config: TeamConfig,
            flo_team: FloTeam,
            router_prompt: ChatPromptTemplate = None,
            llm: Union[BaseLanguageModel, None] = None,
        ) -> None:
            self.name = team_config.router.name
            self.session = session
            self.llm = llm if llm is not None else session.llm
            self.flo_team = flo_team
            self.agents = flo_team.members
            self.members = [agent.name for agent in flo_team.members]
            self.options = self.members + [FLO_FINISH]
            member_type = (
                'workers' if flo_team.members[0].type == 'agent' else 'team members'
            )

            router_base_system_message = (
                'You are a supervisor tasked with managing a conversation between the'
                ' following {member_type}: {members}. Given the following rules,'
                ' respond with the worker to act next '
            )

            self.parser = JsonOutputParser(pydantic_object=NextAgent)
            self.llm_router_prompt = ChatPromptTemplate.from_messages(
                [
                    ('system', router_base_system_message),
                    MessagesPlaceholder(variable_name='messages'),
                    ('system', 'Rules: {router_prompt}'),
                    (
                        'system',
                        'Given the conversation above, who should act next?'
                        ' Or should we FINISH if the task is already answered ? Select one of: {options} \n {format_instructions}',
                    ),
                ]
            ).partial(
                options=str(self.options),
                members=', '.join(self.members),
                member_type=member_type,
                router_prompt=router_prompt,
                format_instructions=self.parser.get_format_instructions(),
            )

        def build(self):
            chain = self.llm_router_prompt | self.llm | self.parser

            return FloLLMRouter(
                executor=chain,
                flo_team=self.flo_team,
                name=self.name,
                session=self.session,
            )



from typing import Optional
from langchain_core.language_models import BaseLanguageModel
from langchain.tools import Tool
from langgraph.graph import END, StateGraph
from langgraph.graph.graph import CompiledGraph
from langgraph.prebuilt import ToolNode
from flo_ai.state.flo_state import TeamFloAgentState
from langgraph.prebuilt import tools_condition
from flo_ai.models.flo_executable import ExecutableFlo
from langchain import hub
from langchain.prompts import ChatPromptTemplate


class FloRag(ExecutableFlo):
    def __init__(self, name: str, graph: CompiledGraph) -> None:
        super().__init__(name, 'team')
        self.graph = graph

    def draw(self, xray=True):
        return self.graph.get_graph(xray=xray).draw_mermaid_png()


class FloRagBuilder:
    def __init__(
        self,
        name: str,
        tools: list[Tool],
        llm: BaseLanguageModel,
        prompt: Optional[ChatPromptTemplate] = None,
    ) -> None:
        self.name = name
        self.llm = llm
        self.tools = tools
        self.prompt = hub.pull('rlm/rag-prompt') if prompt is None else prompt

    def retriever_agent(self, state: TeamFloAgentState):
        messages = state['messages']
        model = self.llm.bind_tools(self.tools)
        response = model.invoke(messages)
        # We return a list, because this will get added to the existing list
        return {'messages': [response]}

    def generate(self, state: TeamFloAgentState):
        messages = state['messages']
        question = messages[0].content
        last_message = messages[-1]

        question = messages[0].content
        docs = last_message.content

        rag_chain = self.prompt | self.llm
        response = rag_chain.invoke({'context': docs, 'question': question})
        return {'messages': [response]}

    def build(self) -> FloRag:
        retrieve = ToolNode(self.tools)

        workflow = StateGraph(TeamFloAgentState)
        workflow.add_node('agent', self.retriever_agent)
        workflow.add_node('retrieve', retrieve)  # retrieval
        workflow.add_node('generate', self.generate)
        workflow.add_conditional_edges(
            'agent',
            tools_condition,
            {
                'tools': 'retrieve',
                END: END,
            },
        )

        workflow.add_edge('retrieve', 'generate')
        workflow.add_edge('generate', END)

        workflow.set_entry_point('agent')
        graph = workflow.compile()
        return FloRag(self.name, graph=graph)



import functools
from typing import Union
from abc import ABC, abstractmethod
from langgraph.graph import END, StateGraph
from flo_ai.models.flo_node import FloNode
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_team import FloTeam
from flo_ai.models.flo_member import FloMember
from flo_ai.yaml.config import TeamConfig
from flo_ai.models.flo_routed_team import FloRoutedTeam
from flo_ai.constants.prompt_constants import FLO_FINISH
from flo_ai.models.flo_executable import ExecutableType
from flo_ai.state.flo_state import (
    TeamFloAgentState,
    STATE_NAME_LOOP_CONTROLLER,
    STATE_NAME_NEXT,
)
from flo_ai.constants.flo_node_contants import (
    INTERNAL_NODE_REFLECTION_MANAGER,
    INTERNAL_NODE_DELEGATION_MANAGER,
)


class FloRouter(ABC):
    def __init__(
        self,
        session: FloSession,
        name: str,
        flo_team: FloTeam,
        executor,
        config: TeamConfig = None,
    ):
        self.router_name = name
        self.session: FloSession = session
        self.flo_team: FloTeam = flo_team
        self.members = flo_team.members
        self.member_names = [x.name for x in flo_team.members]
        self.type: ExecutableType = ExecutableType.router
        self.executor = executor
        self.config = config

    def build_routed_team(self) -> FloRoutedTeam:
        return self.build_graph()

    @abstractmethod
    def build_graph():
        pass

    def build_node(self, flo_member: FloMember) -> FloNode:
        node_builder = FloNode.Builder()
        if flo_member.type == ExecutableType.router:
            return node_builder.build_from_router(flo_member)
        if flo_member.type == ExecutableType.team:
            return node_builder.build_from_team(flo_member)
        if flo_member.type == ExecutableType.delegator:
            return FloNode(
                flo_member.executor, flo_member.name, flo_member.type, flo_member.config
            )
        node_builder = FloNode.Builder()
        return node_builder.build_from_agent(flo_member)

    def router_fn(self, state: TeamFloAgentState):
        next = state['next']
        conditional_map = {k: k for k in self.member_names}
        conditional_map[FLO_FINISH] = END
        self.session.append(node=next)
        if self.session.is_looping(node=next):
            return conditional_map[FLO_FINISH]
        return conditional_map[next]

    def update_reflection_state(
        self, state: TeamFloAgentState, reflection_agent_name: str
    ):
        tracker = state.get(STATE_NAME_LOOP_CONTROLLER) or {}
        tracker[reflection_agent_name] = tracker.get(reflection_agent_name, 0) + 1
        return {STATE_NAME_LOOP_CONTROLLER: tracker}

    def add_delegation_edge(
        self,
        workflow: StateGraph,
        parent: FloNode,
        delegation_node: FloNode,
        nextNode: Union[FloNode | str],
    ):
        to_agent_names = [x.name for x in delegation_node.config.to]
        delegation_node_name = delegation_node.name
        next_node_name = nextNode if isinstance(nextNode, str) else nextNode.name
        retry = delegation_node.config.retry or 1

        conditional_map = {}
        for agent_name in to_agent_names:
            conditional_map[agent_name] = agent_name
        conditional_map[next_node_name] = next_node_name

        workflow.add_node(
            INTERNAL_NODE_DELEGATION_MANAGER,
            functools.partial(
                self.update_reflection_state, reflection_agent_name=delegation_node_name
            ),
        )

        workflow.add_edge(parent.name, INTERNAL_NODE_DELEGATION_MANAGER)
        workflow.add_conditional_edges(
            INTERNAL_NODE_DELEGATION_MANAGER,
            self.__get_refelection_routing_fn(
                retry, delegation_node_name, next_node_name
            ),
            {
                delegation_node_name: delegation_node_name,
                next_node_name: next_node_name,
            },
        )

        workflow.add_conditional_edges(
            delegation_node_name,
            FloRouter.__get_delegation_router_fn(next_node_name),
            conditional_map,
        )

    @staticmethod
    def __get_delegation_router_fn(nextNode: str):
        def delegation_router(state: TeamFloAgentState):
            if STATE_NAME_NEXT not in state:
                return nextNode
            return state[STATE_NAME_NEXT]

        return delegation_router

    def add_reflection_edge(
        self,
        workflow: StateGraph,
        reflection_node: FloNode,
        nextNode: Union[FloNode | str],
    ):
        to_agent_name = reflection_node.config.to[0].name
        retry = reflection_node.config.retry or 1
        reflection_agent_name = reflection_node.name
        next = nextNode if isinstance(nextNode, str) else nextNode.name

        workflow.add_node(
            INTERNAL_NODE_REFLECTION_MANAGER,
            functools.partial(
                self.update_reflection_state,
                reflection_agent_name=reflection_agent_name,
            ),
        )

        workflow.add_edge(to_agent_name, INTERNAL_NODE_REFLECTION_MANAGER)
        workflow.add_conditional_edges(
            INTERNAL_NODE_REFLECTION_MANAGER,
            self.__get_refelection_routing_fn(retry, reflection_agent_name, next),
            {reflection_agent_name: reflection_agent_name, next: next},
        )
        workflow.add_edge(reflection_agent_name, to_agent_name)

    @staticmethod
    def __get_refelection_routing_fn(
        retries: int, reflection_agent_name, next_node_name
    ):
        def reflection_routing_fn(state: TeamFloAgentState):
            tracker = state[STATE_NAME_LOOP_CONTROLLER]
            if (
                tracker is not None
                and reflection_agent_name in tracker
                and tracker[reflection_agent_name] > retries
            ):
                return next_node_name
            return reflection_agent_name

        return reflection_routing_fn



from typing import Optional
from flo_ai.state.flo_session import FloSession
from flo_ai.yaml.config import AgentConfig
from flo_ai.models.flo_agent import FloAgent
from flo_ai.models.flo_llm_agent import FloLLMAgent
from flo_ai.models.flo_reflection_agent import FloReflectionAgent
from flo_ai.models.flo_delegation_agent import FloDelegatorAgent
from flo_ai.models.flo_tool_agent import FloToolAgent
from flo_ai.error.flo_exception import FloException
from flo_ai.constants.common_constants import DOCUMENTATION_AGENT_ANCHOR
from enum import Enum


class AgentKinds(Enum):
    agentic = 'agentic'
    llm = 'llm'
    tool = 'tool'
    function = 'function'
    reflection = 'reflection'
    delegator = 'delegator'


class AgentFactory:
    @staticmethod
    def create(session: FloSession, agent: AgentConfig):
        kind = agent.kind
        tool_map = session.tools
        if kind is not None:
            agent_kind = getattr(AgentKinds, kind, None)
            if agent_kind is None:
                raise FloException(f"""Unknown agent kind: `{kind}`. The supported types are llm, tool, reflection, delegator or agentic. 
                            Check the documentation @ {DOCUMENTATION_AGENT_ANCHOR}""")
            match agent_kind:
                case AgentKinds.llm:
                    return AgentFactory.__create_llm_agent(session, agent)
                case AgentKinds.tool:
                    return AgentFactory.__create_runnable_agent(session, agent)
                case AgentKinds.reflection:
                    return AgentFactory.__create_reflection_agent(session, agent)
                case AgentKinds.delegator:
                    return AgentFactory.__create_delegator_agent(session, agent)
        return AgentFactory.__create_agentic_agent(session, agent, tool_map)

    @staticmethod
    def __resolve_model(session: FloSession, model_name: Optional[str] = None):
        if model_name is None:
            return session.llm
        if model_name not in session.models:
            raise FloException(
                f"""Model not found: {model_name}.
                The model you would like to use should be registered to the session using session.register_model api, 
                and the same model name should be used here instead of `{model_name}`"""
            )
        return session.models[model_name]

    @staticmethod
    def __create_agentic_agent(
        session: FloSession, agent: AgentConfig, tool_map
    ) -> FloAgent:
        agent_model = AgentFactory.__resolve_model(session, agent.model)
        tools = [tool_map[tool.name] for tool in agent.tools]
        flo_agent: FloAgent = FloAgent.Builder(
            session, agent, tools, llm=agent_model, on_error=session.on_agent_error
        ).build()
        return flo_agent

    @staticmethod
    def __create_llm_agent(session: FloSession, agent: AgentConfig) -> FloLLMAgent:
        agent_model = AgentFactory.__resolve_model(session, agent.model)
        builder = FloLLMAgent.Builder(session, agent, llm=agent_model)
        llm_agent: FloLLMAgent = builder.build()
        return llm_agent

    @staticmethod
    def __create_runnable_agent(session: FloSession, agent: AgentConfig) -> FloLLMAgent:
        runnable = session.tools[agent.tools[0].name]
        return FloToolAgent.Builder(session, agent, runnable).build()

    @staticmethod
    def __create_reflection_agent(
        session: FloSession, agent: AgentConfig
    ) -> FloReflectionAgent:
        agent_model = AgentFactory.__resolve_model(session, agent.model)
        return FloReflectionAgent.Builder(session, agent, llm=agent_model).build()

    @staticmethod
    def __create_delegator_agent(
        session: FloSession, agent: AgentConfig
    ) -> FloReflectionAgent:
        agent_model = AgentFactory.__resolve_model(session, agent.model)
        return FloDelegatorAgent.Builder(session, agent, llm=agent_model).build()



from langchain_core.runnables import Runnable
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.state.flo_session import FloSession
from flo_ai.yaml.config import AgentConfig
from flo_ai.models.flo_executable import ExecutableType


class FloToolAgent(ExecutableFlo):
    def __init__(self, executor: Runnable, config: AgentConfig) -> None:
        super().__init__(config.name, executor, ExecutableType.tool)
        self.executor: Runnable = executor
        self.config: AgentConfig = config

    class Builder:
        def __init__(
            self, session: FloSession, config: AgentConfig, tool_runnable: Runnable
        ) -> None:
            self.name: str = config.name
            self.runnable = tool_runnable
            self.config = config

        def build(self) -> Runnable:
            return FloToolAgent(self.runnable, self.config)



import asyncio
import warnings
import logging
from typing import Optional
from flo_ai.yaml.config import to_supervised_team
from flo_ai.builders.yaml_builder import build_supervised_team, FloRoutedTeamConfig
from typing import Any, Iterator, Union
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.error.flo_exception import FloException
from flo_ai.constants.common_constants import DOCUMENTATION_WEBSITE
from flo_ai.common.flo_logger import (
    get_logger,
    set_log_level_internal,
    set_log_config_internal,
    set_logger_internal,
    FloLogConfig,
)


class Flo:
    def __init__(self, session: FloSession, config: FloRoutedTeamConfig) -> None:
        self.session = session
        self.config = config
        session.config = config
        self.runnable: ExecutableFlo = build_supervised_team(session)

        self.langchain_logger = session.langchain_logger
        get_logger().info('Flo instance created ...', session)

    def stream(self, query, config=None) -> Iterator[Union[dict[str, Any], Any]]:
        self.validate_invoke(self.session)
        get_logger().info(f"streaming query requested: '{query}'", self.session)
        return self.runnable.stream(query, config)

    def async_stream(self, query, config=None) -> Iterator[Union[dict[str, Any], Any]]:
        get_logger().info(f"Streaming async query requested: '{query}'", self.session)
        return self.runnable.astream(query, config)

    def invoke(self, query, config=None) -> Iterator[Union[dict[str, Any], Any]]:
        config = {'callbacks': [self.session.langchain_logger]}
        self.validate_invoke(self.session)
        get_logger().info(f"Invoking query: '{query}'", self.session)
        return self.runnable.invoke(query, config)

    def async_invoke(self, query, config=None) -> Iterator[Union[dict[str, Any], Any]]:
        config = {'callbacks': [self.session.langchain_logger]}
        get_logger().info(f"Invoking async query: '{query}'", self.session)
        return self.runnable.ainvoke(query, config)

    @staticmethod
    def build(session: FloSession, yaml: str, log_level: Optional[str] = None):
        if log_level:
            warnings.warn(
                '`log_level` is deprecated and will be removed in a future version. '
                'Please use `Flo.set_log_level()` instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            Flo.set_log_level(log_level)
        get_logger().info('Building Flo instance from YAML ...', session)
        return Flo(session, to_supervised_team(yaml))

    @staticmethod
    def set_log_level(log_level: str):
        set_log_level_internal(log_level)

    @staticmethod
    def set_log_config(logging_config: FloLogConfig):
        set_log_config_internal(logging_config)

    @staticmethod
    def set_logger(logging_config: logging.Logger):
        set_logger_internal(logging_config)

    def draw(self, xray=True):
        from IPython.display import Image, display

        return display(Image(self.runnable.draw(xray)))

    def draw_to_file(self, filename: str, xray=True):
        from PIL import Image as PILImage
        import io

        byte_image = self.runnable.draw(xray)
        with io.BytesIO(byte_image) as image_io:
            image = PILImage.open(image_io)
            image.save(filename)

    def validate_invoke(self, session: FloSession):
        async_coroutines = filter(
            lambda x: (
                hasattr(x, 'coroutine') and asyncio.iscoroutinefunction(x.coroutine)
            ),
            session.tools.values(),
        )
        if len(list(async_coroutines)) > 0:
            raise FloException(
                f"""You seem to have atleast one async tool registered in this session. Please use flo.async_invoke or flo.async_stream. Checkout {DOCUMENTATION_WEBSITE}"""
            )



from flo_ai.models.flo_member import FloMember
from flo_ai.yaml.config import TeamConfig


class FloTeam:
    def __init__(self, team_config: TeamConfig, members: list[FloMember]) -> None:
        self.name = team_config.name
        self.config = team_config
        self.members = members

    class Builder:
        def __init__(self, team_config: TeamConfig, members: list[FloMember]) -> None:
            self.team_config = team_config
            self.members = members
            self.member_names = list(map(lambda x: x.name, self.members))

        def build(self):
            return FloTeam(team_config=self.team_config, members=self.members)



from typing import Optional
from langchain_core.runnables import Runnable
from flo_ai.yaml.config import AgentConfig
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableFlo
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from flo_ai.models.flo_executable import ExecutableType
from langchain_core.output_parsers.openai_functions import JsonOutputFunctionsParser
from langchain_core.language_models import BaseLanguageModel


class FloDelegatorAgent(ExecutableFlo):
    def __init__(self, executor: Runnable, config: AgentConfig) -> None:
        super().__init__(config.name, executor, ExecutableType.delegator)
        self.executor: Runnable = executor
        self.config: AgentConfig = config

    class Builder:
        def __init__(
            self,
            session: FloSession,
            agentConfig: AgentConfig,
            llm: Optional[BaseLanguageModel] = None,
        ) -> None:
            self.config = agentConfig
            delegator_base_system_message = (
                'You are a delegator tasked with routing a conversation between the'
                ' following {member_type}: {members}. Given the following rules,'
                ' respond with the worker to act next '
            )
            self.llm = session.llm if llm is None else llm
            self.options = [x.name for x in agentConfig.to]
            self.llm_router_prompt = ChatPromptTemplate.from_messages(
                [
                    ('system', delegator_base_system_message),
                    MessagesPlaceholder(variable_name='messages'),
                    ('system', 'Rules: {delegator_rules}'),
                    (
                        'system',
                        'Given the conversation above, who should act next?'
                        'Select one of: {options}',
                    ),
                ]
            ).partial(
                options=str(self.options),
                members=', '.join(self.options),
                member_type='agents',
                delegator_rules=agentConfig.job,
            )

        def build(self):
            function_def = {
                'name': 'route',
                'description': 'Select the next role.',
                'parameters': {
                    'title': 'routeSchema',
                    'type': 'object',
                    'properties': {
                        'next': {
                            'title': 'Next',
                            'anyOf': [
                                {'enum': self.options},
                            ],
                        }
                    },
                    'required': ['next'],
                },
            }

            chain = (
                self.llm_router_prompt
                | self.llm.bind_functions(
                    functions=[function_def], function_call='route'
                )
                | JsonOutputFunctionsParser()
            )

            return FloDelegatorAgent(executor=chain, config=self.config)



from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from typing import Union
from langchain_core.runnables import Runnable
from flo_ai.state.flo_session import FloSession
from flo_ai.constants.prompt_constants import FLO_FINISH
from flo_ai.router.flo_llm_router import FloLLMRouter
from flo_ai.models.flo_team import FloTeam
from flo_ai.yaml.config import TeamConfig
from langchain_core.output_parsers import JsonOutputParser
from flo_ai.router.flo_llm_router import NextAgent

supervisor_system_message = (
    'You are a supervisor tasked with managing a conversation between the'
    ' following {member_type}: {members}. Given the following user request,'
    ' respond with the worker to act next. Each worker will perform a'
    ' task and respond with their results and status. When the users question is answered or the assigned task is finished,'
    ' respond with FINISH. '
)


class FloSupervisor(FloLLMRouter):
    def __init__(
        self, session: FloSession, executor: Runnable, flo_team: FloTeam, name: str
    ) -> None:
        super().__init__(
            session=session, name=name, flo_team=flo_team, executor=executor
        )

    class Builder:
        def __init__(
            self,
            session: FloSession,
            team_config: TeamConfig,
            flo_team: FloTeam,
            llm: Union[BaseLanguageModel, None] = None,
        ) -> None:
            self.name = team_config.router.name
            self.session = session
            self.llm = llm if llm is not None else session.llm
            self.flo_team = flo_team
            self.agents = flo_team.members
            self.members = [agent.name for agent in flo_team.members]
            self.options = self.members + [FLO_FINISH]
            member_type = (
                'workers' if flo_team.members[0].type == 'agent' else 'team members'
            )
            self.parser = JsonOutputParser(pydantic_object=NextAgent)
            self.supervisor_prompt = ChatPromptTemplate.from_messages(
                [
                    ('system', supervisor_system_message),
                    MessagesPlaceholder(variable_name='messages'),
                    (
                        'system',
                        'Given the conversation above, who should act next?'
                        ' Or should we FINISH if the task is already answered, Select one of: {options}  \n {format_instructions}',
                    ),
                ]
            ).partial(
                options=str(self.options),
                members=', '.join(self.members),
                member_type=member_type,
                format_instructions=self.parser.get_format_instructions(),
            )

        def build(self):
            chain = self.supervisor_prompt | self.llm | self.parser

            return FloSupervisor(
                executor=chain,
                flo_team=self.flo_team,
                name=self.name,
                session=self.session,
            )



from typing import Union
from langchain_core.runnables import Runnable
from flo_ai.yaml.config import AgentConfig
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableFlo
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from flo_ai.models.flo_executable import ExecutableType
from langchain_core.output_parsers import StrOutputParser


class FloReflectionAgent(ExecutableFlo):
    def __init__(self, executor: Runnable, config: AgentConfig) -> None:
        super().__init__(config.name, executor, ExecutableType.reflection)
        self.config = config

    class Builder:
        def __init__(
            self,
            session: FloSession,
            config: AgentConfig,
            llm: Union[BaseLanguageModel, None] = None,
        ) -> None:
            prompt_message: Union[ChatPromptTemplate, str] = config.job
            self.name: str = config.name
            self.llm = llm if llm is not None else session.llm
            self.config = config

            system_prompts = (
                [
                    ('system', 'You are a {}'.format(config.role)),
                    ('system', prompt_message),
                ]
                if config.role is not None
                else [('system', prompt_message)]
            )
            system_prompts.append(MessagesPlaceholder(variable_name='messages'))
            self.prompt: ChatPromptTemplate = (
                ChatPromptTemplate.from_messages(system_prompts)
                if isinstance(prompt_message, str)
                else prompt_message
            )

        def build(self):
            executor = self.prompt | self.llm | StrOutputParser()
            return FloReflectionAgent(executor, self.config)



import functools
from flo_ai.models.flo_agent import FloAgent
from flo_ai.models.flo_routed_team import FloRoutedTeam
from langchain.agents import AgentExecutor
from flo_ai.state.flo_state import TeamFloAgentState, STATE_NAME_MESSAGES
from langchain_core.messages import HumanMessage
from flo_ai.yaml.config import AgentConfig, TeamConfig
from flo_ai.models.flo_executable import ExecutableType
from typing import Union


class FloNode:
    def __init__(
        self,
        func: functools.partial,
        name: str,
        kind: ExecutableType,
        config: Union[AgentConfig | TeamConfig],
    ) -> None:
        self.name = name
        self.func = func
        self.kind: ExecutableType = kind
        self.config: Union[AgentConfig | TeamConfig] = config

    class Builder:
        def build_from_agent(self, flo_agent: FloAgent) -> 'FloNode':
            agent_func = functools.partial(
                FloNode.Builder.__teamflo_agent_node,
                agent=flo_agent.runnable,
                name=flo_agent.name,
                agent_config=flo_agent.config,
            )
            return FloNode(agent_func, flo_agent.name, flo_agent.type, flo_agent.config)

        def build_from_team(self, flo_team: FloRoutedTeam) -> 'FloNode':
            team_chain = (
                functools.partial(
                    FloNode.Builder.__teamflo_team_node, members=flo_team.runnable.nodes
                )
                | flo_team.runnable
            )
            return FloNode(
                (
                    FloNode.Builder.__get_last_message
                    | team_chain
                    | FloNode.Builder.__join_graph
                ),
                flo_team.name,
                flo_team.type,
                flo_team.config,
            )

        def build_from_router(self, flo_router) -> 'FloNode':
            router_func = functools.partial(
                FloNode.Builder.__teamflo_router_node,
                agent=flo_router.executor,
                name=flo_router.router_name,
                agent_config=flo_router.config,
            )
            return FloNode(
                router_func, flo_router.router_name, flo_router.type, flo_router.config
            )

        @staticmethod
        def __teamflo_agent_node(
            state: TeamFloAgentState,
            agent: AgentExecutor,
            name: str,
            agent_config: AgentConfig,
        ):
            result = agent.invoke(state)
            output = result if isinstance(result, str) else result['output']
            return {STATE_NAME_MESSAGES: [HumanMessage(content=output, name=name)]}

        @staticmethod
        def __teamflo_router_node(
            state: TeamFloAgentState,
            agent: AgentExecutor,
            name: str,
            agent_config: AgentConfig,
        ):
            result = agent.invoke(state)
            nextNode = result if isinstance(result, str) else result['next']
            return {'next': nextNode}

        @staticmethod
        def __get_last_message(state: TeamFloAgentState) -> str:
            return state[STATE_NAME_MESSAGES][-1].content

        @staticmethod
        def __join_graph(response: dict):
            return {STATE_NAME_MESSAGES: [response[STATE_NAME_MESSAGES][-1]]}

        @staticmethod
        def __teamflo_team_node(message: str, members: list[str]):
            results = {
                STATE_NAME_MESSAGES: [HumanMessage(content=message)],
                'team_members': ', '.join(members),
            }
            return results



DOCUMENTATION_WEBSITE = 'https://flo-ai.rootflo.ai'
DOCUMENTATION_ROUTER_ANCHOR = f'{DOCUMENTATION_WEBSITE}/basics/routers'
DOCUMENTATION_AGENT_ANCHOR = f'{DOCUMENTATION_WEBSITE}/basics/agents'
DOCUMENTATION_AGENT_TOOLS = f'{DOCUMENTATION_WEBSITE}/basics/tools'



import warnings
from typing import Union, Dict
from langchain_core.language_models import BaseLanguageModel
from langchain_core.tools import BaseTool
from flo_ai.common.flo_logger import get_logger
from flo_ai.common.flo_langchain_logger import FloLangchainLogger
from flo_ai.yaml.config import FloRoutedTeamConfig, FloAgentConfig
from flo_ai.helpers.utils import random_str

from typing import Optional


def _handle_agent_error(error) -> str:
    error_message = str(error)[:50]
    return f"""
            Following error happened while agent execution, please retry with the fix for the same:
            {error_message}
        """


class FloSession:
    def __init__(
        self,
        default_llm: BaseLanguageModel = None,
        loop_size: int = 2,
        max_loop: int = 3,
        llm: BaseLanguageModel = None,
        log_level: Optional[str] = None,
        callbacks: Optional[FloLangchainLogger] = None,
        on_agent_error=_handle_agent_error,
    ) -> None:
        if log_level:
            warnings.warn(
                '`log_level` is deprecated and will be removed in a future version. '
                'Please use `Flo.set_log_level()` instead.',
                DeprecationWarning,
                stacklevel=2,
            )

        self.session_id = str(random_str(16))
        self.llm = self.resolve_llm(default_llm, llm)
        self.tools = dict()
        self.models: Dict[str, BaseLanguageModel] = dict()
        self.tools: Dict[str, BaseTool] = dict()
        self.counter = dict()
        self.navigation: list[str] = list()
        self.pattern_series = dict()
        self.loop_size: int = loop_size
        self.max_loop: int = max_loop
        self.on_agent_error = on_agent_error

        self.config: Union[FloRoutedTeamConfig, FloAgentConfig] = None
        get_logger().info('New session created ...', self)
        self.langchain_logger = FloLangchainLogger(self.session_id)

    def resolve_llm(
        self, default_llm: BaseLanguageModel = None, llm: BaseLanguageModel = None
    ):
        if default_llm is not None:
            return default_llm
        if llm:
            warnings.warn(
                '`llm` is deprecated and will be removed in a future version. '
                'Please use `default_llm` instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            return llm

    def register_tool(self, name: str, tool: BaseTool):
        self.tools[name] = tool
        get_logger().info(f"Tool '{name}' registered for session {self.session_id}")
        return self

    def register_model(self, name: str, model: BaseLanguageModel):
        self.models[name] = model
        get_logger().info(f"Model '{name}' registered for session {self.session_id}")
        return self

    def append(self, node: str) -> int:
        get_logger().debug(f'Appending node: {node}')
        self.counter[node] = self.counter.get(node, 0) + 1
        if node in self.navigation:
            last_known_index = (
                len(self.navigation) - 1 - self.navigation[::-1].index(node)
            )
            pattern_array = self.navigation[last_known_index : len(self.navigation)]
            if len(pattern_array) + 1 >= self.loop_size:
                pattern = '|'.join(pattern_array) + '|' + node
                if node in self.pattern_series:
                    self.pattern_series[node].append(pattern)
                else:
                    self.pattern_series[node] = [pattern]
        self.navigation.append(node)

    def is_looping(self, node) -> bool:
        get_logger().debug(f'Checking if node {node} is looping')
        patterns = self.pattern_series[node] if node in self.pattern_series else []
        if len(patterns) < self.max_loop:
            return False
        return patterns[-(self.max_loop) :] == [patterns[-1]] * self.max_loop

    def stringify(self):
        return str(self.counter)



from flo_ai.models.flo_executable import ExecutableFlo
from langgraph.graph.graph import CompiledGraph
from flo_ai.yaml.config import TeamConfig


class FloRoutedTeam(ExecutableFlo):
    def __init__(self, name: str, graph: CompiledGraph, config: TeamConfig) -> None:
        super().__init__(name, graph)
        self.config = config

    # Overridden for xray use, doesnt work in base class
    def draw(self, xray=True):
        return self.runnable.get_graph(xray=xray).draw_mermaid_png()



from flo_ai.core import Flo as Flo
from flo_ai.models.flo_agent import FloAgent as FloAgent
from flo_ai.router.flo_supervisor import FloSupervisor as FloSupervisor
from flo_ai.models.flo_team import FloTeam as FloTeam
from flo_ai.router.flo_linear import FloLinear as FloLinear
from flo_ai.state.flo_session import FloSession as FloSession
from flo_ai.retrievers.flo_retriever import FloRagBuilder as FloRagBuilder
from flo_ai.common.flo_logger import get_logger as get_logger
from flo_ai.common.flo_langchain_logger import FloLangchainLogger as FloLangchainLogger



from langchain_core.embeddings import Embeddings
from langchain_text_splitters import CharacterTextSplitter
from langchain_community.document_transformers import EmbeddingsRedundantFilter
from langchain.retrievers.document_compressors import EmbeddingsFilter


class FloCompressionPipeline:
    def __init__(self, embeddings: Embeddings) -> None:
        self.__embeddings = embeddings
        self.__pipeline = []

    def add_chuncking(self, chunk_size=300, chunk_overlap=0):
        splitter = CharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap, separator='. '
        )
        self.__pipeline.append(splitter)

    def add_embedding_reduntant_filter(self):
        redundant_filter = EmbeddingsRedundantFilter(embeddings=self.__embeddings)
        self.__pipeline.append(redundant_filter)

    def add_embedding_relevant_filter(self, threshold: float = 0.50):
        relevant_filter = EmbeddingsFilter(
            embeddings=self.__embeddings, similarity_threshold=threshold
        )
        self.__pipeline.append(relevant_filter)

    def add_flashrank_reranking(self, model_name='ms-marco-MultiBERT-L-12'):
        from langchain.retrievers.document_compressors.flashrank_rerank import (
            FlashrankRerank,
        )

        compressor = FlashrankRerank(model=model_name)
        self.__pipeline.append(compressor)

    def add_cohere_reranking(self, model_name='rerank-english-v3.0'):
        from langchain.retrievers.document_compressors.cohere_rerank import CohereRerank

        compressor = CohereRerank(model=model_name)
        self.__pipeline.append(compressor)

    def get(self):
        return self.__pipeline



from flo_ai.yaml.config import TeamConfig
from flo_ai.router.flo_router import FloRouter
from langgraph.graph import StateGraph, END, START
from flo_ai.state.flo_state import TeamFloAgentState
from flo_ai.models.flo_routed_team import FloRoutedTeam
from flo_ai.models.flo_team import FloTeam
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableType


class FloLinear(FloRouter):
    def __init__(self, session: FloSession, config: TeamConfig, flo_team: FloTeam):
        super().__init__(
            session=session,
            name=config.name,
            flo_team=flo_team,
            executor=None,
            config=config,
        )
        self.router_config = config.router

    def build_graph(self):
        flo_agent_nodes = [self.build_node(member) for member in self.members]

        workflow = StateGraph(TeamFloAgentState)

        for flo_node in flo_agent_nodes:
            agent_name = flo_node.name
            workflow.add_node(agent_name, flo_node.func)

        if self.router_config.edges is None:
            start_node = flo_agent_nodes[0]
            end_node = flo_agent_nodes[-1]
            workflow.add_edge(START, start_node.name)
            for i in range(len(flo_agent_nodes) - 1):
                parent_node = flo_agent_nodes[i]
                child_node = flo_agent_nodes[i + 1]
                next_node = (
                    flo_agent_nodes[i + 2] if (i + 2) < len(flo_agent_nodes) else END
                )
                if parent_node.kind == ExecutableType.reflection:
                    self.add_reflection_edge(workflow, parent_node, child_node)
                    continue
                if child_node.kind == ExecutableType.delegator:
                    self.add_delegation_edge(
                        workflow, parent_node, child_node, next_node
                    )
                    continue

                if (
                    child_node.kind != ExecutableType.reflection
                    and parent_node.kind != ExecutableType.delegator
                ):
                    workflow.add_edge(parent_node.name, child_node.name)

            if end_node.kind == ExecutableType.reflection:
                self.add_reflection_edge(workflow, end_node, END)
            elif end_node.kind != ExecutableType.delegator:
                workflow.add_edge(end_node.name, END)
        else:
            workflow.add_edge(START, self.router_config.start_node)
            for edge in self.router_config.edges:
                workflow.add_edge(edge[0], edge[1])
            workflow.add_edge(self.router_config.end_node, END)

        workflow_graph = workflow.compile()

        return FloRoutedTeam(self.flo_team.name, workflow_graph, self.flo_team.config)

    class Builder:
        def __init__(
            self, session: FloSession, config: TeamConfig, flo_team: FloTeam
        ) -> None:
            self.config = config
            self.session = session
            self.team = flo_team

        def build(self):
            return FloLinear(self.session, self.config, self.team)
